@@ -1,14 +1,10 @@
 """Top-level package for rigid-body-motion."""
 __author__ = """Peter Hausamann"""
 __email__ = "peter.hausamann@tum.de"
-__version__ = "0.1.0"
-from pathlib import Path
+__version__ = "0.7.0"
 
-from pkg_resources import resource_filename
-
-from . import ros  # noqa
+from . import io, plot, ros  # noqa
 from .coordinate_systems import (
-    _replace_dim,
     cartesian_to_polar,
     cartesian_to_spherical,
     polar_to_cartesian,
@@ -18,19 +14,42 @@ from .core import (
     _make_dataarray,
     _make_transform_or_pose_dataset,
     _make_twist_dataset,
+    _make_velocity_dataarray,
     _maybe_unpack_dataarray,
+    _replace_dim,
     _resolve_rf,
     _transform,
 )
 from .estimators import (
+    best_fit_rotation,
     best_fit_transform,
+    estimate_angular_velocity,
+    estimate_linear_velocity,
     iterative_closest_point,
     shortest_arc_rotation,
 )
 from .reference_frames import ReferenceFrame
 from .reference_frames import _registry as registry
-from .reference_frames import clear_registry, deregister_frame, register_frame
-from .utils import qinv, qmean, rotate_vectors
+from .reference_frames import (
+    clear_registry,
+    deregister_frame,
+    register_frame,
+    render_tree,
+)
+from .utils import (
+    ExampleDataStore,
+    from_euler_angles,
+    qinterp,
+    qinv,
+    qmean,
+    qmul,
+    rotate_vectors,
+)
+
+try:
+    import rigid_body_motion.accessors  # noqa
+except ImportError:
+    pass
 
 __all__ = [
     "transform_points",
@@ -49,16 +68,26 @@ __all__ = [
     "deregister_frame",
     "clear_registry",
     "ReferenceFrame",
+    "render_tree",
     # estimators
+    "estimate_linear_velocity",
+    "estimate_angular_velocity",
     "shortest_arc_rotation",
+    "best_fit_rotation",
     "best_fit_transform",
     "iterative_closest_point",
     "lookup_transform",
+    "lookup_pose",
     "lookup_twist",
+    "lookup_linear_velocity",
+    "lookup_angular_velocity",
     # utils
+    "from_euler_angles",
     "example_data",
+    "qinterp",
     "qinv",
     "qmean",
+    "qmul",
     "rotate_vectors",
 ]
 
@@ -72,19 +101,18 @@ _cs_funcs = {
 }
 
 
-example_data = {
-    "head": Path(resource_filename("rigid_body_motion", "data/head.nc")),
-    "left_eye": Path(
-        resource_filename("rigid_body_motion", "data/left_eye.nc")
-    ),
-    "right_eye": Path(
-        resource_filename("rigid_body_motion", "data/right_eye.nc")
-    ),
-}
+example_data = ExampleDataStore()
 
 
 def transform_vectors(
-    arr, into, outof=None, dim=None, axis=None, timestamps=None, time_axis=None
+    arr,
+    into,
+    outof=None,
+    dim=None,
+    axis=None,
+    timestamps=None,
+    time_axis=None,
+    return_timestamps=False,
 ):
     """ Transform an array of vectors between reference frames.
 
@@ -122,6 +150,9 @@ def transform_vectors(
         The axis of the array representing the timestamps of the vectors.
         Defaults to the first axis of the array.
 
+    return_timestamps: bool, default False
+        If True, also return the timestamps after the transformation.
+
     Returns
     -------
     arr_transformed: array_like
@@ -135,12 +166,27 @@ def transform_vectors(
     transform_quaternions, transform_points, ReferenceFrame
     """
     return _transform(
-        "transform_vectors", arr, into, outof, dim, axis, timestamps, time_axis
+        "transform_vectors",
+        arr,
+        into,
+        outof,
+        dim,
+        axis,
+        timestamps,
+        time_axis,
+        return_timestamps=return_timestamps,
     )
 
 
 def transform_points(
-    arr, into, outof=None, dim=None, axis=None, timestamps=None, time_axis=None
+    arr,
+    into,
+    outof=None,
+    dim=None,
+    axis=None,
+    timestamps=None,
+    time_axis=None,
+    return_timestamps=False,
 ):
     """ Transform an array of points between reference frames.
 
@@ -154,9 +200,9 @@ def transform_points(
         which the array will be represented after the transformation.
 
     outof: str or ReferenceFrame, optional
-        ReferenceFrame instance or name of a registered reference frame in
-        which the array is currently represented. Can be omitted if the array
-        is a DataArray whose ``attrs`` contain a "representation_frame" entry
+        ReferenceFrame instance or name of a registered reference frame which
+        is the current reference frame of the array. Can be omitted if the
+        array is a DataArray whose ``attrs`` contain a "reference_frame" entry
         with the name of a registered frame.
 
     dim: str, optional
@@ -178,6 +224,9 @@ def transform_points(
         The axis of the array representing the timestamps of the points.
         Defaults to the first axis of the array.
 
+    return_timestamps: bool, default False
+        If True, also return the timestamps after the transformation.
+
     Returns
     -------
     arr_transformed: array_like
@@ -191,12 +240,27 @@ def transform_points(
     transform_vectors, transform_quaternions, ReferenceFrame
     """
     return _transform(
-        "transform_points", arr, into, outof, dim, axis, timestamps, time_axis
+        "transform_points",
+        arr,
+        into,
+        outof,
+        dim,
+        axis,
+        timestamps,
+        time_axis,
+        return_timestamps=return_timestamps,
     )
 
 
 def transform_quaternions(
-    arr, into, outof=None, dim=None, axis=None, timestamps=None, time_axis=None
+    arr,
+    into,
+    outof=None,
+    dim=None,
+    axis=None,
+    timestamps=None,
+    time_axis=None,
+    return_timestamps=False,
 ):
     """ Transform an array of quaternions between reference frames.
 
@@ -210,9 +274,9 @@ def transform_quaternions(
         which the array will be represented after the transformation.
 
     outof: str or ReferenceFrame, optional
-        ReferenceFrame instance or name of a registered reference frame in
-        which the array is currently represented. Can be omitted if the array
-        is a DataArray whose ``attrs`` contain a "representation_frame" entry
+        ReferenceFrame instance or name of a registered reference frame which
+        is the current reference frame of the array. Can be omitted if the
+        array is a DataArray whose ``attrs`` contain a "reference_frame" entry
         with the name of a registered frame.
 
     dim: str, optional
@@ -234,6 +298,9 @@ def transform_quaternions(
         The axis of the array representing the timestamps of the quaternions.
         Defaults to the first axis of the array.
 
+    return_timestamps: bool, default False
+        If True, also return the timestamps after the transformation.
+
     Returns
     -------
     arr_transformed: array_like
@@ -255,6 +322,7 @@ def transform_quaternions(
         axis,
         timestamps,
         time_axis,
+        return_timestamps=return_timestamps,
     )
 
 
@@ -268,6 +336,7 @@ def transform_angular_velocity(
     timestamps=None,
     time_axis=None,
     cutoff=None,
+    return_timestamps=False,
 ):
     """ Transform an array of angular velocities between frames.
 
@@ -280,6 +349,18 @@ def transform_angular_velocity(
     the frame that is being changed and will be represented in the new
     frame after the transformation.
 
+    When transforming the reference frame R to a new frame R' while keeping
+    the moving frame M fixed, the transformed velocity is calculated
+    according to the formula:
+
+    .. math:: \omega_{M/R'} = \omega_{M/R} + \omega_{R/R'}
+
+    When transforming the moving frame M to a new frame M' while keeping
+    the reference frame R fixed, the transformed velocity is calculated
+    according to the formula:
+
+    .. math:: \omega_{M'/R} = \omega_{M/R} + \omega_{M'/M}
+
     Parameters
     ----------
     arr: array_like
@@ -322,6 +403,9 @@ def transform_angular_velocity(
         velocity after the twist estimation as a fraction of the Nyquist
         frequency.
 
+    return_timestamps: bool, default False
+        If True, also return the timestamps after the transformation.
+
     Returns
     -------
     arr_transformed: array_like
@@ -334,7 +418,7 @@ def transform_angular_velocity(
     --------
     transform_linear_velocity, transform_vectors, transform_quaternions,
     transform_points, ReferenceFrame
-    """
+    """  # noqa
     return _transform(
         "transform_angular_velocity",
         arr,
@@ -346,6 +430,7 @@ def transform_angular_velocity(
         time_axis,
         what=what,
         cutoff=cutoff,
+        return_timestamps=return_timestamps,
     )
 
 
@@ -362,6 +447,7 @@ def transform_linear_velocity(
     time_axis=None,
     cutoff=None,
     outlier_thresh=None,
+    return_timestamps=False,
 ):
     """ Transform an array of linear velocities between frames.
 
@@ -374,6 +460,18 @@ def transform_linear_velocity(
     the frame that is being changed and will be represented in the new
     frame after the transformation.
 
+    When transforming the reference frame R to a new frame R' while keeping
+    the moving frame M fixed, the transformed velocity is calculated
+    according to the formula:
+
+    .. math:: v_{M/R'} = v_{M/R} + v_{R/R'} + \omega_{R/R'} \\times t_{M/R}
+
+    When transforming the moving frame M to a new frame M' while keeping
+    the reference frame R fixed, the transformed velocity is calculated
+    according to the formula:
+
+    .. math:: v_{M'/R} = v_{M/R} + v_{M'/M} + \omega_{M/R} \\times t_{M'/M}
+
     Parameters
     ----------
     arr: array_like
@@ -434,6 +532,9 @@ def transform_linear_velocity(
         measurements from the Intel RealSense T265 tracker, set this value
         to 1e-3.
 
+    return_timestamps: bool, default False
+        If True, also return the timestamps after the transformation.
+
     Returns
     -------
     arr_transformed: array_like
@@ -446,7 +547,7 @@ def transform_linear_velocity(
     --------
     transform_angular_velocity, transform_vectors, transform_quaternions,
     transform_points, ReferenceFrame
-    """
+    """  # noqa
     return _transform(
         "transform_linear_velocity",
         arr,
@@ -461,6 +562,7 @@ def transform_linear_velocity(
         reference_frame=reference_frame,
         cutoff=cutoff,
         outlier_thresh=outlier_thresh,
+        return_timestamps=return_timestamps,
     )
 
 
@@ -510,7 +612,7 @@ def transform_coordinates(
     spherical_to_cartesian
     """
     arr, axis, _, _, _, _, coords, dims, name, attrs = _maybe_unpack_dataarray(
-        arr, dim, axis
+        arr, dim, axis, timestamps=False
     )
 
     if outof is None:
@@ -545,9 +647,15 @@ def transform_coordinates(
         return arr
 
 
-def lookup_transform(outof, into, as_dataset=False):
+def lookup_transform(outof, into, as_dataset=False, return_timestamps=False):
     """ Look up transformation from one frame to another.
 
+    The transformation is a rotation `r` followed by a translation `t` which,
+    when applied to a point expressed wrt the base frame `B`, yields that
+    point wrt the target frame `T`:
+
+    .. math:: p_T = rot(r, p_B) + t
+
     Parameters
     ----------
     outof: str or ReferenceFrame
@@ -558,42 +666,98 @@ def lookup_transform(outof, into, as_dataset=False):
 
     as_dataset: bool, default False
         If True, return an xarray.Dataset. Otherwise, return a tuple of
-        translation, rotation and timestamps.
+        translation and rotation.
+
+    return_timestamps: bool, default False
+        If True, and `as_dataset` is False, also return the timestamps of the
+        lookup.
 
     Returns
     -------
-    translation, rotation, timestamps: each numpy.ndarray
-        Translation, rotation and timestamps of transformation between the
-        frames, if `as_dataset` is False.
+    translation, rotation: each numpy.ndarray
+        Translation and rotation of transformation between the frames,
+        if `as_dataset` is False.
+
+    timestamps: numpy.ndarray
+        Corresponding timestamps of the lookup if `return_timestamps` is True.
 
     ds: xarray.Dataset
         The above arrays as an xarray.Dataset, if `as_dataset` is True.
     """
     into = _resolve_rf(into)
     outof = _resolve_rf(outof)
-    translation, rotation, timestamps = into.get_transformation(outof)
+    translation, rotation, timestamps = outof.lookup_transform(into)
 
     if as_dataset:
         return _make_transform_or_pose_dataset(
             translation, rotation, outof, timestamps
         )
-    else:
+    elif return_timestamps:
         return translation, rotation, timestamps
+    else:
+        return translation, rotation
+
+
+def lookup_pose(frame, reference, as_dataset=False, return_timestamps=False):
+    """ Look up pose of one frame wrt a reference.
+
+    Parameters
+    ----------
+    frame: str or ReferenceFrame
+        Frame for which to look up the pose.
+
+    reference: str or ReferenceFrame
+        Reference frame of the pose.
+
+    as_dataset: bool, default False
+        If True, return an xarray.Dataset. Otherwise, return a tuple of
+        position and orientation.
+
+    return_timestamps: bool, default False
+        If True, and `as_dataset` is False, also return the timestamps of the
+        lookup.
+
+    Returns
+    -------
+    position, orientation: each numpy.ndarray
+        Position and orientation of the pose between the frames,
+        if `as_dataset` is False.
+
+    timestamps: numpy.ndarray
+        Corresponding timestamps of the lookup if `return_timestamps` is True.
+
+    ds: xarray.Dataset
+        The above arrays as an xarray.Dataset, if `as_dataset` is True.
+    """
+    reference = _resolve_rf(reference)
+    frame = _resolve_rf(frame)
+    position, orientation, timestamps = frame.lookup_transform(reference)
+
+    if as_dataset:
+        return _make_transform_or_pose_dataset(
+            position, orientation, reference, timestamps, pose=True
+        )
+    elif return_timestamps:
+        return position, orientation, timestamps
+    else:
+        return position, orientation
 
 
 def lookup_twist(
-    moving_frame,
+    frame,
     reference=None,
     represent_in=None,
     outlier_thresh=None,
     cutoff=None,
+    mode="quaternion",
     as_dataset=False,
+    return_timestamps=False,
 ):
     """ Estimate linear and angular velocity of a frame wrt a reference.
 
     Parameters
     ----------
-    moving_frame: str or ReferenceFrame
+    frame: str or ReferenceFrame
         The reference frame whose twist is estimated.
 
     reference: str or ReferenceFrame, optional
@@ -619,31 +783,204 @@ def lookup_twist(
         velocity after the estimation as a fraction of the Nyquist
         frequency.
 
+    mode: str, default "quaternion"
+        If "quaternion", compute the angular velocity from the quaternion
+        derivative. If "rotation_vector", compute the angular velocity from
+        the gradient of the axis-angle representation of the rotations.
+
     as_dataset: bool, default False
         If True, return an xarray.Dataset. Otherwise, return a tuple of linear
-        and angular velocity and timestamps.
+        and angular velocity.
+
+    return_timestamps: bool, default False
+        If True, and `as_dataset` is False, also return the timestamps of the
+        lookup.
 
     Returns
     -------
-    linear, angular, timestamps: each numpy.ndarray
-        Linear and angular velocity and timestamps of moving frame wrt
-        reference frame, represented in representation frame, if `as_dataset`
-        is False.
+    linear, angular: each numpy.ndarray
+        Linear and angular velocity of moving frame wrt reference frame,
+        represented in representation frame, if `as_dataset` is False.
+
+    timestamps: numpy.ndarray
+        Corresponding timestamps of the lookup if `return_timestamps` is True.
 
     ds: xarray.Dataset
         The above arrays as an xarray.Dataset, if `as_dataset` is True.
     """
-    moving_frame = _resolve_rf(moving_frame)
-    reference = _resolve_rf(reference or moving_frame.parent)
+    frame = _resolve_rf(frame)
+    reference = _resolve_rf(reference or frame.parent)
     represent_in = _resolve_rf(represent_in or reference)
 
-    linear, angular, timestamps = moving_frame.lookup_twist(
-        reference, represent_in, outlier_thresh=outlier_thresh, cutoff=cutoff
+    linear, angular, timestamps = frame.lookup_twist(
+        reference,
+        represent_in,
+        outlier_thresh=outlier_thresh,
+        cutoff=cutoff,
+        mode=mode,
+        return_timestamps=True,
     )
 
     if as_dataset:
         return _make_twist_dataset(
-            angular, linear, moving_frame, reference, represent_in, timestamps
+            angular, linear, frame, reference, represent_in, timestamps
         )
-    else:
+    elif return_timestamps:
         return linear, angular, timestamps
+    else:
+        return linear, angular
+
+
+def lookup_linear_velocity(
+    frame,
+    reference=None,
+    represent_in=None,
+    outlier_thresh=None,
+    cutoff=None,
+    as_dataarray=False,
+    return_timestamps=False,
+):
+    """ Estimate linear velocity of a frame wrt a reference.
+
+    Parameters
+    ----------
+    frame: str or ReferenceFrame
+        The reference frame whose velocity is estimated.
+
+    reference: str or ReferenceFrame, optional
+        The reference frame wrt which the velocity is estimated. Defaults to
+        the parent frame of the moving frame.
+
+    represent_in: str or ReferenceFrame, optional
+        The reference frame in which the twist is represented. Defaults
+        to the reference frame.
+
+    outlier_thresh: float, optional
+        Some SLAM-based trackers introduce position corrections when a new
+        camera frame becomes available. This introduces outliers in the
+        linear velocity estimate. The estimation algorithm used here
+        can suppress these outliers by throwing out samples where the
+        norm of the second-order differences of the position is above
+        `outlier_thresh` and interpolating the missing values. For
+        measurements from the Intel RealSense T265 tracker, set this value
+        to 1e-3.
+
+    cutoff: float, optional
+        Frequency of a low-pass filter applied to linear and angular
+        velocity after the estimation as a fraction of the Nyquist
+        frequency.
+
+    as_dataarray: bool, default False
+        If True, return an xarray.DataArray.
+
+    return_timestamps: bool, default False
+        If True and `as_dataarray` is False, also return the timestamps of the
+        lookup.
+
+    Returns
+    -------
+    linear: numpy.ndarray or xarray.DataArray
+        Linear velocity of moving frame wrt reference frame, represented in
+        representation frame.
+
+    timestamps: numpy.ndarray
+        Corresponding timestamps of the lookup if `return_timestamps` is True.
+    """
+    frame = _resolve_rf(frame)
+    reference = _resolve_rf(reference or frame.parent)
+    represent_in = _resolve_rf(represent_in or reference)
+
+    linear, timestamps = frame.lookup_linear_velocity(
+        reference,
+        represent_in,
+        outlier_thresh=outlier_thresh,
+        cutoff=cutoff,
+        return_timestamps=True,
+    )
+
+    if as_dataarray:
+        return _make_velocity_dataarray(
+            linear, "linear", frame, reference, represent_in, timestamps
+        )
+    elif return_timestamps:
+        return linear, timestamps
+    else:
+        return linear
+
+
+def lookup_angular_velocity(
+    frame,
+    reference=None,
+    represent_in=None,
+    outlier_thresh=None,
+    cutoff=None,
+    mode="quaternion",
+    as_dataarray=False,
+    return_timestamps=False,
+):
+    """ Estimate angular velocity of a frame wrt a reference.
+
+    Parameters
+    ----------
+    frame: str or ReferenceFrame
+        The reference frame whose velocity is estimated.
+
+    reference: str or ReferenceFrame, optional
+        The reference frame wrt which the velocity is estimated. Defaults to
+        the parent frame of the moving frame.
+
+    represent_in: str or ReferenceFrame, optional
+        The reference frame in which the twist is represented. Defaults
+        to the reference frame.
+
+    outlier_thresh: float, optional
+        Suppress samples where the norm of the second-order differences of the
+        rotation is above `outlier_thresh` and interpolate the missing values.
+
+    cutoff: float, optional
+        Frequency of a low-pass filter applied to angular and angular
+        velocity after the estimation as a fraction of the Nyquist
+        frequency.
+
+    mode: str, default "quaternion"
+        If "quaternion", compute the angular velocity from the quaternion
+        derivative. If "rotation_vector", compute the angular velocity from
+        the gradient of the axis-angle representation of the rotations.
+
+    as_dataarray: bool, default False
+        If True, return an xarray.DataArray.
+
+    return_timestamps: bool, default False
+        If True and `as_dataarray` is False, also return the timestamps of the
+        lookup.
+
+    Returns
+    -------
+    angular: numpy.ndarray or xarray.DataArray
+        Angular velocity of moving frame wrt reference frame, represented in
+        representation frame.
+
+    timestamps: numpy.ndarray
+        Corresponding timestamps of the lookup if `return_timestamps` is True.
+    """
+    frame = _resolve_rf(frame)
+    reference = _resolve_rf(reference or frame.parent)
+    represent_in = _resolve_rf(represent_in or reference)
+
+    angular, timestamps = frame.lookup_angular_velocity(
+        reference,
+        represent_in,
+        outlier_thresh=outlier_thresh,
+        cutoff=cutoff,
+        mode=mode,
+        return_timestamps=True,
+    )
+
+    if as_dataarray:
+        return _make_velocity_dataarray(
+            angular, "angular", frame, reference, represent_in, timestamps,
+        )
+    elif return_timestamps:
+        return angular, timestamps
+    else:
+        return angular
