@@ -1,16 +1,14 @@
 """"""
 import numpy as np
-from anytree import NodeMixin, Walker
-from quaternion import (
-    as_float_array,
-    as_quat_array,
-    derivative,
-    from_rotation_matrix,
-)
-from scipy.interpolate import interp1d
-from scipy.signal import butter, filtfilt
+from anytree import NodeMixin, RenderTree, Walker
+from quaternion import as_float_array, as_quat_array, from_rotation_matrix
 
-from rigid_body_motion.core import _resolve_rf
+from rigid_body_motion.core import (
+    TransformMatcher,
+    _estimate_angular_velocity,
+    _estimate_linear_velocity,
+    _resolve_rf,
+)
 from rigid_body_motion.utils import qinv, rotate_vectors
 
 _registry = {}
@@ -43,6 +41,18 @@ def _deregister(name):
     _registry.pop(name)
 
 
+def render_tree(root):
+    """ Render a reference frame tree.
+
+    Parameters
+    ----------
+    root: str or ReferenceFrame
+        The root of the rendered tree.
+    """
+    for pre, _, node in RenderTree(_resolve_rf(root)):
+        print(f"{pre}{node.name}")
+
+
 def register_frame(
     name,
     parent=None,
@@ -50,6 +60,7 @@ def register_frame(
     rotation=None,
     timestamps=None,
     inverse=False,
+    discrete=False,
     update=False,
 ):
     """ Register a new reference frame in the registry.
@@ -81,6 +92,11 @@ def register_frame(
         translation and rotation are specified for the parent frame wrt this
         frame.
 
+    discrete: bool, default False
+        If True, transformations with timestamps are assumed to be events.
+        Instead of interpolating between timestamps, transformations are
+        fixed between their timestamp and the next one.
+
     update: bool, default False
         If True, overwrite if there is a frame with the same name in the
         registry.
@@ -93,6 +109,7 @@ def register_frame(
         rotation=rotation,
         timestamps=timestamps,
         inverse=inverse,
+        discrete=discrete,
     )
     _register(rf, update=update)
 
@@ -124,6 +141,7 @@ class ReferenceFrame(NodeMixin):
         rotation=None,
         timestamps=None,
         inverse=False,
+        discrete=False,
     ):
         """ Constructor.
 
@@ -145,7 +163,7 @@ class ReferenceFrame(NodeMixin):
             The rotation of this frame wrt the parent frame. Not
             applicable if there is no parent frame.
 
-        timestamps : array_like, optional
+        timestamps: array_like, optional
             The timestamps for translation and rotation of this frame. Not
             applicable if this is a static reference frame.
 
@@ -153,6 +171,11 @@ class ReferenceFrame(NodeMixin):
             If True, invert the transform wrt the parent frame, i.e. the
             translation and rotation are specified for the parent frame wrt
             this frame.
+
+        discrete: bool, default False
+            If True, transformations with timestamps are assumed to be events.
+            Instead of interpolating between timestamps, transformations are
+            fixed between their timestamp and the next one.
         """
         super(ReferenceFrame, self).__init__()
 
@@ -171,6 +194,11 @@ class ReferenceFrame(NodeMixin):
             self._verify_root(translation, rotation, timestamps)
             self.translation, self.rotation, self.timestamps = None, None, None
 
+        if discrete and self.timestamps is None:
+            raise ValueError("timestamps must be provided when discrete=True")
+        else:
+            self.discrete = discrete
+
     def __del__(self):
         """ Destructor. """
         if self.name in _registry and _registry[self.name] is self:
@@ -235,121 +263,6 @@ class ReferenceFrame(NodeMixin):
         if timestamps is not None:
             raise ValueError("timestamps specified without parent frame.")
 
-    @classmethod
-    def _broadcast(cls, arr, timestamps):
-        """ Broadcast scalar array along timestamp axis. """
-        # TODO test
-        return np.tile(arr, (len(timestamps), 1))
-
-    @classmethod
-    def _interpolate(cls, source_arr, target_arr, source_ts, target_ts):
-        """ Interpolate source array at target array timestamps. """
-        # TODO SLERP for quaternions
-        # TODO specify time_axis as parameter
-        # TODO priority=None/<rf_name>
-        # TODO method + optional scipy dependency?
-        ts_dtype = target_ts.dtype
-        source_ts = source_ts.astype(float)
-        target_ts = target_ts.astype(float)
-
-        # TODO sort somewhere and turn these into assertions or use min/max
-        #  with boolean indexing
-        if np.any(np.diff(source_ts) < 0):
-            raise ValueError("source_ts is not sorted.")
-        if np.any(np.diff(target_ts) < 0):
-            raise ValueError("target_ts is not sorted.")
-
-        # TODO raise error when intersection is empty
-        if target_ts[0] < source_ts[0]:
-            target_arr = target_arr[target_ts >= source_ts[0]]
-            target_ts = target_ts[target_ts >= source_ts[0]]
-        if target_ts[-1] > source_ts[-1]:
-            target_arr = target_arr[target_ts <= source_ts[-1]]
-            target_ts = target_ts[target_ts <= source_ts[-1]]
-
-        source_arr_interp = interp1d(source_ts, source_arr, axis=0)(target_ts)
-
-        return source_arr_interp, target_arr, target_ts.astype(ts_dtype)
-
-    @classmethod
-    def _match_timestamps(cls, arr, arr_ts, rf_t, rf_r, rf_ts):
-        """ Match timestamps of array and reference frame. """
-        # TODO test
-        # TODO policy='from_arr'/'from_rf'
-        if rf_ts is None:
-            return arr, rf_t, rf_r, arr_ts
-        elif arr_ts is None:
-            return cls._broadcast(arr, rf_ts), rf_t, rf_r, rf_ts
-        elif len(arr_ts) != len(rf_ts) or np.any(arr_ts != rf_ts):
-            # abuse interpolate by stacking t and r and splitting afterwards
-            rf_tr = np.hstack((rf_t, rf_r))
-            arr, rf_tr, rf_ts = cls._interpolate(arr, rf_tr, arr_ts, rf_ts)
-            rf_t, rf_r = np.hsplit(rf_tr, [3])
-            return arr, rf_t, rf_r, rf_ts
-        else:
-            return arr, rf_t, rf_r, rf_ts
-
-    @classmethod
-    def _match_timestamps_multi(cls, arr_list, ts_list):
-        """ Match multiple arrays and timestamps at once. """
-        earliest = np.max([ts[0] for ts in ts_list if ts is not None])
-        latest = np.min([ts[-1] for ts in ts_list if ts is not None])
-
-        return_ts = None
-        for ts in ts_list:
-            if return_ts is None and ts is not None:
-                return_ts = ts[(ts >= earliest) & (ts <= latest)]
-                break
-        else:
-            return arr_list, None
-
-        return_arr_list = []
-        for arr, ts in zip(arr_list, ts_list):
-            if ts is None:
-                return_arr_list.append(cls._broadcast(arr, return_ts))
-            else:
-                return_arr_list.append(
-                    interp1d(ts.astype(float), arr, axis=0)(
-                        return_ts.astype(float)
-                    )
-                )
-
-        return return_arr_list, return_ts
-
-    @classmethod
-    def _add_transformation(cls, rf, t, r, ts, inverse=False):
-        """ Add transformation of this frame to current transformation. """
-        # TODO test
-        if rf.timestamps is not None:
-            if ts is None:
-                translation = rf.translation
-                rotation = rf.rotation
-                t = cls._broadcast(t, rf.timestamps)
-                r = cls._broadcast(r, rf.timestamps)
-                ts_new = rf.timestamps
-            else:
-                translation, t, ts_new = cls._interpolate(
-                    rf.translation, t, rf.timestamps, ts
-                )
-                rotation, r, ts_new = cls._interpolate(
-                    rf.rotation, r, rf.timestamps, ts
-                )
-        else:
-            translation = rf.translation
-            rotation = rf.rotation
-            ts_new = ts
-
-        if inverse:
-            q = 1 / as_quat_array(rotation)
-            dt = -np.array(translation)
-            t = rotate_vectors(q, t + dt)
-        else:
-            q = as_quat_array(rotation)
-            dt = np.array(translation)
-            t = rotate_vectors(q, t) + dt
-
-        return t, as_float_array(q * as_quat_array(r)), ts_new
-
     @classmethod
     def _validate_input(cls, arr, axis, n_axis, timestamps, time_axis):
         """ Validate shape of array and timestamps. """
@@ -371,6 +284,7 @@ class ReferenceFrame(NodeMixin):
                     f"Axis {time_axis} of the array must have the same length "
                     f"as the timestamps"
                 )
+            # TODO this should be done somewhere else
             arr = np.swapaxes(arr, 0, time_axis)
 
         return arr, timestamps
@@ -384,6 +298,16 @@ class ReferenceFrame(NodeMixin):
 
         return t_or_r
 
+    @classmethod
+    def _match_arrays(cls, arrays, timestamps=None):
+        """ Match multiple arrays with timestamps. """
+        matcher = TransformMatcher()
+
+        for array in arrays:
+            matcher.add_array(*array)
+
+        return matcher.get_arrays(timestamps)
+
     def _walk(self, to_rf):
         """ Walk from this frame to a target frame along the tree. """
         to_rf = _resolve_rf(to_rf)
@@ -391,6 +315,22 @@ class ReferenceFrame(NodeMixin):
         up, _, down = walker.walk(self, to_rf)
         return up, down
 
+    def _get_matcher(self, to_frame, arrays=None):
+        """ Get a TransformMatcher from this frame to another. """
+        up, down = self._walk(to_frame)
+
+        matcher = TransformMatcher()
+        for rf in up:
+            matcher.add_reference_frame(rf)
+        for rf in down:
+            matcher.add_reference_frame(rf, inverse=True)
+
+        if arrays is not None:
+            for array in arrays:
+                matcher.add_array(*array)
+
+        return matcher
+
     @classmethod
     def from_dataset(
         cls,
@@ -401,6 +341,7 @@ class ReferenceFrame(NodeMixin):
         parent,
         name=None,
         inverse=False,
+        discrete=False,
     ):
         """ Construct a reference frame from a Dataset.
 
@@ -433,6 +374,11 @@ class ReferenceFrame(NodeMixin):
             translation and rotation are specified for the parent frame wrt
             this frame.
 
+        discrete: bool, default False
+            If True, transformations with timestamps are assumed to be events.
+            Instead of interpolating between timestamps, transformations are
+            fixed between their timestamp and the next one.
+
         Returns
         -------
         rf: ReferenceFrame
@@ -446,11 +392,12 @@ class ReferenceFrame(NodeMixin):
             ds[rotation].data,
             ds[timestamps].data,
             inverse=inverse,
+            discrete=discrete,
         )
 
     @classmethod
     def from_translation_dataarray(
-        cls, da, timestamps, parent, name=None, inverse=False
+        cls, da, timestamps, parent, name=None, inverse=False, discrete=False,
     ):
         """ Construct a reference frame from a translation DataArray.
 
@@ -475,6 +422,11 @@ class ReferenceFrame(NodeMixin):
             If True, invert the transform wrt the parent frame, i.e. the
             translation is specified for the parent frame wrt this frame.
 
+        discrete: bool, default False
+            If True, transformations with timestamps are assumed to be events.
+            Instead of interpolating between timestamps, transformations are
+            fixed between their timestamp and the next one.
+
         Returns
         -------
         rf: ReferenceFrame
@@ -487,11 +439,12 @@ class ReferenceFrame(NodeMixin):
             translation=da.data,
             timestamps=da[timestamps].data,
             inverse=inverse,
+            discrete=discrete,
         )
 
     @classmethod
     def from_rotation_dataarray(
-        cls, da, timestamps, parent, name=None, inverse=False
+        cls, da, timestamps, parent, name=None, inverse=False, discrete=False,
     ):
         """ Construct a reference frame from a rotation DataArray.
 
@@ -516,6 +469,11 @@ class ReferenceFrame(NodeMixin):
             If True, invert the transform wrt the parent frame, i.e. the
             rotation is specified for the parent frame wrt this frame.
 
+        discrete: bool, default False
+            If True, transformations with timestamps are assumed to be events.
+            Instead of interpolating between timestamps, transformations are
+            fixed between their timestamp and the next one.
+
         Returns
         -------
         rf: ReferenceFrame
@@ -528,6 +486,7 @@ class ReferenceFrame(NodeMixin):
             rotation=da.data,
             timestamps=da[timestamps].data,
             inverse=inverse,
+            discrete=discrete,
         )
 
     @classmethod
@@ -570,12 +529,25 @@ class ReferenceFrame(NodeMixin):
         )
 
     def get_transformation(self, to_frame):
-        """ Calculate the transformation from this frame to another.
+        """ Alias for lookup_transform.
 
-        The transformation is a rotation followed by a translation which,
-        when applied to a position and/or orientation represented in this
-        reference frame, yields the representation of that
-        position/orientation in the target reference frame.
+        See Also
+        --------
+        ReferenceFrame.lookup_transform
+        """
+        import warnings
+
+        warnings.warn(
+            DeprecationWarning(
+                "get_transformation is deprecated, use lookup_transform "
+                "instead."
+            )
+        )
+
+        return self.lookup_transform(to_frame)
+
+    def lookup_transform(self, to_frame):
+        """ Look up the transformation from this frame to another.
 
         Parameters
         ----------
@@ -593,20 +565,14 @@ class ReferenceFrame(NodeMixin):
 
         ts: array_like, shape (n_timestamps,) or None
             The timestamps for which the transformation is defined.
+
+        See Also
+        --------
+        lookup_transform
         """
-        up, down = self._walk(to_frame)
+        matcher = self._get_matcher(to_frame)
 
-        t = np.zeros(3)
-        r = np.array((1.0, 0.0, 0.0, 0.0))
-        ts = None
-
-        for rf in up:
-            t, r, ts = self._add_transformation(rf, t, r, ts)
-
-        for rf in down:
-            t, r, ts = self._add_transformation(rf, t, r, ts, inverse=True)
-
-        return t, r, ts
+        return matcher.get_transformation()
 
     def transform_vectors(
         self,
@@ -653,10 +619,10 @@ class ReferenceFrame(NodeMixin):
             The timestamps after the transformation.
         """
         arr, arr_ts = self._validate_input(arr, axis, 3, timestamps, time_axis)
+        matcher = self._get_matcher(to_frame, arrays=[(arr, arr_ts)])
+        t, r, ts = matcher.get_transformation()
+        arr, _ = matcher.get_arrays(ts)
 
-        t, r, rf_ts = self.get_transformation(to_frame)
-
-        arr, _, r, ts = self._match_timestamps(arr, arr_ts, t, r, rf_ts)
         r = self._expand_singleton_axes(r, arr.ndim)
         arr = rotate_vectors(r, arr, axis=axis)
 
@@ -714,10 +680,10 @@ class ReferenceFrame(NodeMixin):
             The timestamps after the transformation.
         """
         arr, arr_ts = self._validate_input(arr, axis, 3, timestamps, time_axis)
+        matcher = self._get_matcher(to_frame, arrays=[(arr, arr_ts)])
+        t, r, ts = matcher.get_transformation()
+        arr, _ = matcher.get_arrays(ts)
 
-        t, r, rf_ts = self.get_transformation(to_frame)
-
-        arr, t, r, ts = self._match_timestamps(arr, arr_ts, t, r, rf_ts)
         t = self._expand_singleton_axes(t, arr.ndim)
         r = self._expand_singleton_axes(r, arr.ndim)
         arr = rotate_vectors(r, arr, axis=axis)
@@ -778,10 +744,10 @@ class ReferenceFrame(NodeMixin):
             The timestamps after the transformation.
         """
         arr, arr_ts = self._validate_input(arr, axis, 4, timestamps, time_axis)
+        matcher = self._get_matcher(to_frame, arrays=[(arr, arr_ts)])
+        t, r, ts = matcher.get_transformation()
+        arr, _ = matcher.get_arrays(ts)
 
-        t, r, rf_ts = self.get_transformation(to_frame)
-
-        arr, _, r, ts = self._match_timestamps(arr, arr_ts, t, r, rf_ts)
         r = self._expand_singleton_axes(r, arr.ndim)
         arr = np.swapaxes(arr, axis, -1)
         arr = as_quat_array(r) * as_quat_array(arr)
@@ -809,15 +775,6 @@ class ReferenceFrame(NodeMixin):
     ):
         """ Transform array of angular velocities from this frame to another.
 
-        The array represents the velocity of a moving body or frame wrt a
-        reference frame, expressed in a representation frame.
-
-        The transformation changes either the reference frame, the moving
-        frame or the representation frame of the velocity from this frame to
-        another. In either case, it is assumed that the array is represented in
-        the frame that is being changed and will be represented in the new
-        frame after the transformation.
-
         Parameters
         ----------
         arr: array_like
@@ -860,15 +817,31 @@ class ReferenceFrame(NodeMixin):
 
         ts: array_like, shape (n_timestamps,) or None
             The timestamps after the transformation.
+
+        See Also
+        --------
+        transform_angular_velocity
         """
         if what == "reference_frame":
-            _, angular, angular_ts = self.lookup_twist(
-                to_frame, to_frame, cutoff=cutoff
+            angular, angular_ts = self.lookup_angular_velocity(
+                to_frame,
+                to_frame,
+                cutoff=cutoff,
+                allow_static=True,
+                return_timestamps=True,
             )
+
         elif what == "moving_frame":
-            _, angular, angular_ts = _resolve_rf(to_frame).lookup_twist(
-                self, to_frame, cutoff=cutoff
+            angular, angular_ts = _resolve_rf(
+                to_frame
+            ).lookup_angular_velocity(
+                self,
+                to_frame,
+                cutoff=cutoff,
+                allow_static=True,
+                return_timestamps=True,
             )
+
         elif what == "representation_frame":
             return self.transform_vectors(
                 arr,
@@ -878,10 +851,11 @@ class ReferenceFrame(NodeMixin):
                 timestamps=timestamps,
                 return_timestamps=return_timestamps,
             )
+
         else:
             raise ValueError(
-                f"Expected 'what' to be 'reference_frame' or 'moving_frame', "
-                f"got {what}"
+                f"Expected 'what' to be 'reference_frame', 'moving_frame' or "
+                f"'representation_frame', got {what}"
             )
 
         arr, ts = self.transform_vectors(
@@ -893,14 +867,13 @@ class ReferenceFrame(NodeMixin):
             return_timestamps=True,
         )
 
-        angular, arr, timestamps = self._interpolate(
-            angular, arr, angular_ts, ts
+        arr, angular, ts_out = self._match_arrays(
+            [(arr, ts), (angular, angular_ts)]
         )
-
         arr += angular
 
         if return_timestamps:
-            return arr, timestamps
+            return arr, ts_out
         else:
             return arr
 
@@ -920,15 +893,6 @@ class ReferenceFrame(NodeMixin):
     ):
         """ Transform array of linear velocities from this frame to another.
 
-        The array represents the velocity of a moving body or frame wrt a
-        reference frame, expressed in a representation frame.
-
-        The transformation changes either the reference frame, the moving
-        frame or the representation frame of the velocity from this frame to
-        another. In either case, it is assumed that the array is represented in
-        the frame that is being changed and will be represented in the new
-        frame after the transformation.
-
         Parameters
         ----------
         arr: array_like
@@ -973,14 +937,9 @@ class ReferenceFrame(NodeMixin):
             frequency.
 
         outlier_thresh: float, optional
-            Some SLAM-based trackers introduce position corrections when a new
-            camera frame becomes available. This introduces outliers in the
-            linear velocity estimate. The estimation algorithm used here
-            can suppress these outliers by throwing out samples where the
+            Suppress outliers by throwing out samples where the
             norm of the second-order differences of the position is above
-            `outlier_thresh` and interpolating the missing values. For
-            measurements from the Intel RealSense T265 tracker, set this value
-            to 1e-3.
+            `outlier_thresh` and interpolating the missing values.
 
         Returns
         -------
@@ -989,6 +948,10 @@ class ReferenceFrame(NodeMixin):
 
         ts: array_like, shape (n_timestamps,) or None
             The timestamps after the transformation.
+
+        See Also
+        --------
+        transform_linear_velocity
         """
         if what == "reference_frame":
             linear, angular, linear_ts = self.lookup_twist(
@@ -996,21 +959,32 @@ class ReferenceFrame(NodeMixin):
                 to_frame,
                 cutoff=cutoff,
                 outlier_thresh=outlier_thresh,
+                allow_static=True,
+                return_timestamps=True,
             )
             angular_ts = linear_ts
             translation, _, translation_ts = _resolve_rf(
                 moving_frame
-            ).get_transformation(self)
+            ).lookup_transform(self)
 
         elif what == "moving_frame":
             to_frame = _resolve_rf(to_frame)
-            linear, _, linear_ts = to_frame.lookup_twist(
-                self, to_frame, cutoff=cutoff, outlier_thresh=outlier_thresh,
+            linear, linear_ts = to_frame.lookup_linear_velocity(
+                self,
+                to_frame,
+                cutoff=cutoff,
+                outlier_thresh=outlier_thresh,
+                allow_static=True,
+                return_timestamps=True,
             )
-            _, angular, angular_ts = self.lookup_twist(
-                reference_frame, to_frame, cutoff=cutoff
+            angular, angular_ts = self.lookup_angular_velocity(
+                reference_frame,
+                to_frame,
+                cutoff=cutoff,
+                allow_static=True,
+                return_timestamps=True,
             )
-            translation, _, translation_ts = to_frame.get_transformation(self)
+            translation, _, translation_ts = to_frame.lookup_transform(self)
 
         elif what == "representation_frame":
             return self.transform_vectors(
@@ -1024,8 +998,8 @@ class ReferenceFrame(NodeMixin):
 
         else:
             raise ValueError(
-                f"Expected 'what' to be 'reference_frame' or 'moving_frame', "
-                f"got {what}"
+                f"Expected 'what' to be 'reference_frame', 'moving_frame' or "
+                f"'representation_frame', got {what}"
             )
 
         arr, ts = self.transform_vectors(
@@ -1043,63 +1017,30 @@ class ReferenceFrame(NodeMixin):
             return_timestamps=True,
         )
 
-        (arr, linear, angular, translation), ts = self._match_timestamps_multi(
-            [arr, linear, angular, translation],
-            [ts, linear_ts, angular_ts, translation_ts],
+        arr, linear, angular, translation, ts_out = self._match_arrays(
+            [
+                (arr, ts),
+                (linear, linear_ts),
+                (angular, angular_ts),
+                (translation, translation_ts),
+            ]
         )
-
         arr = arr + linear + np.cross(angular, translation)
 
         if return_timestamps:
-            return arr, ts
+            return arr, ts_out
         else:
             return arr
 
-    @classmethod
-    def _estimate_linear_velocity(
-        cls, translation, timestamps, outlier_thresh=None, cutoff=None
-    ):
-        """ Estimate linear velocity of transform. """
-        timestamps = timestamps.astype(float) / 1e9
-
-        linear = np.gradient(translation, timestamps, axis=0)
-
-        if outlier_thresh is not None:
-            dt = np.linalg.norm(np.diff(translation, n=2, axis=0), axis=1)
-            dt = np.hstack((dt, 0.0, 0.0)) + np.hstack((0.0, dt, 0.0))
-            linear = interp1d(
-                timestamps[dt <= outlier_thresh],
-                linear[dt <= outlier_thresh],
-                axis=0,
-                bounds_error=False,
-            )(timestamps)
-
-        if cutoff is not None:
-            linear = filtfilt(*butter(7, cutoff), linear, axis=0)
-
-        return linear
-
-    @classmethod
-    def _estimate_angular_velocity(cls, rotation, timestamps, cutoff=None):
-        """ Estimate angular velocity of transform. """
-        timestamps = timestamps.astype(float) / 1e9
-
-        dq = derivative(rotation, timestamps)
-        angular = as_float_array(
-            2 * as_quat_array(rotation).conjugate() * as_quat_array(dq)
-        )[:, 1:]
-
-        if cutoff is not None:
-            angular = filtfilt(*butter(7, cutoff), angular, axis=0)
-
-        return angular
-
     def lookup_twist(
         self,
         reference=None,
         represent_in=None,
         outlier_thresh=None,
         cutoff=None,
+        mode="quaternion",
+        allow_static=False,
+        return_timestamps=False,
     ):
         """ Estimate linear and angular velocity of this frame wrt a reference.
 
@@ -1114,20 +1055,28 @@ class ReferenceFrame(NodeMixin):
             to the parent frame.
 
         outlier_thresh: float, optional
-            Some SLAM-based trackers introduce position corrections when a new
-            camera frame becomes available. This introduces outliers in the
-            linear velocity estimate. The estimation algorithm used here
-            can suppress these outliers by throwing out samples where the
+            Suppress outliers by throwing out samples where the
             norm of the second-order differences of the position is above
-            `outlier_thresh` and interpolating the missing values. For
-            measurements from the Intel RealSense T265 tracker, set this value
-            to 1e-3.
+            `outlier_thresh` and interpolating the missing values.
 
         cutoff: float, optional
             Frequency of a low-pass filter applied to linear and angular
             velocity after the estimation as a fraction of the Nyquist
             frequency.
 
+        mode: str, default "quaternion"
+            If "quaternion", compute the angular velocity from the quaternion
+            derivative. If "rotation_vector", compute the angular velocity from
+            the gradient of the axis-angle representation of the rotations.
+
+        allow_static: bool, default False
+            If True, return a zero velocity vector and None for timestamps if
+            the transform between this frame and the reference frame is static.
+            Otherwise, a `ValueError` will be raised.
+
+        return_timestamps: bool, default False
+            If True, also return the timestamps of the lookup.
+
         Returns
         -------
         linear: numpy.ndarray, shape (N, 3)
@@ -1147,30 +1096,216 @@ class ReferenceFrame(NodeMixin):
         except TypeError:
             raise ValueError(f"Frame {self.name} has no parent frame")
 
-        translation, rotation, timestamps = self.get_transformation(reference)
+        translation, rotation, timestamps = self.lookup_transform(reference)
 
         if timestamps is None:
-            raise ValueError("Twist cannot be estimated for static transforms")
+            if allow_static:
+                return np.zeros(3), np.zeros(3), None
+            else:
+                raise ValueError(
+                    "Twist cannot be estimated for static transforms"
+                )
 
-        linear = self._estimate_linear_velocity(
-            translation, timestamps, outlier_thresh, cutoff
+        linear = _estimate_linear_velocity(
+            translation,
+            timestamps,
+            outlier_thresh=outlier_thresh,
+            cutoff=cutoff,
+        )
+        angular = _estimate_angular_velocity(
+            rotation, timestamps, cutoff=cutoff, mode=mode
         )
-        angular = self._estimate_angular_velocity(rotation, timestamps, cutoff)
 
+        # linear velocity is represented in reference frame after estimation
         linear, linear_ts = reference.transform_vectors(
             linear, represent_in, timestamps=timestamps, return_timestamps=True
         )
+        # angular velocity is represented in moving frame after estimation
         angular, angular_ts = self.transform_vectors(
             angular,
             represent_in,
             timestamps=timestamps,
             return_timestamps=True,
         )
-        angular, linear, timestamps = self._interpolate(
-            angular, linear, angular_ts, linear_ts
+        angular, linear, twist_ts = self._match_arrays(
+            [(angular, angular_ts), (linear, linear_ts)],
         )
 
-        return linear, angular, timestamps
+        if return_timestamps:
+            return linear, angular, twist_ts
+        else:
+            return linear, angular
+
+    def lookup_linear_velocity(
+        self,
+        reference=None,
+        represent_in=None,
+        outlier_thresh=None,
+        cutoff=None,
+        allow_static=False,
+        return_timestamps=False,
+    ):
+        """ Estimate linear velocity of this frame wrt a reference.
+
+        Parameters
+        ----------
+        reference: str or ReferenceFrame, optional
+            The reference frame wrt which the twist is estimated. Defaults to
+            the parent frame.
+
+        represent_in: str or ReferenceFrame, optional
+            The reference frame in which the twist is represented. Defaults
+            to the parent frame.
+
+        outlier_thresh: float, optional
+            Suppress outliers by throwing out samples where the
+            norm of the second-order differences of the position is above
+            `outlier_thresh` and interpolating the missing values.
+
+        cutoff: float, optional
+            Frequency of a low-pass filter applied to linear and angular
+            velocity after the estimation as a fraction of the Nyquist
+            frequency.
+
+        allow_static: bool, default False
+            If True, return a zero velocity vector and None for timestamps if
+            the transform between this frame and the reference frame is static.
+            Otherwise, a `ValueError` will be raised.
+
+        return_timestamps: bool, default False
+            If True, also return the timestamps of the lookup.
+
+        Returns
+        -------
+        linear: numpy.ndarray, shape (N, 3)
+            Linear velocity of moving frame wrt reference frame, represented
+            in representation frame.
+
+        timestamps: each numpy.ndarray
+            Timestamps of the linear velocity.
+        """
+        try:
+            reference = _resolve_rf(reference or self.parent)
+            represent_in = _resolve_rf(represent_in or self.parent)
+        except TypeError:
+            raise ValueError(f"Frame {self.name} has no parent frame")
+
+        translation, _, timestamps = self.lookup_transform(reference)
+
+        if timestamps is None:
+            if allow_static:
+                return np.zeros(3), None
+            else:
+                raise ValueError(
+                    "Velocity cannot be estimated for static transforms"
+                )
+
+        linear = _estimate_linear_velocity(
+            translation,
+            timestamps,
+            outlier_thresh=outlier_thresh,
+            cutoff=cutoff,
+        )
+        # linear velocity is represented in reference frame after estimation
+        linear, linear_ts = reference.transform_vectors(
+            linear, represent_in, timestamps=timestamps, return_timestamps=True
+        )
+
+        if return_timestamps:
+            return linear, linear_ts
+        else:
+            return linear
+
+    def lookup_angular_velocity(
+        self,
+        reference=None,
+        represent_in=None,
+        outlier_thresh=None,
+        cutoff=None,
+        mode="quaternion",
+        allow_static=False,
+        return_timestamps=False,
+    ):
+        """ Estimate angular velocity of this frame wrt a reference.
+
+        Parameters
+        ----------
+        reference: str or ReferenceFrame, optional
+            The reference frame wrt which the twist is estimated. Defaults to
+            the parent frame.
+
+        represent_in: str or ReferenceFrame, optional
+            The reference frame in which the twist is represented. Defaults
+            to the parent frame.
+
+        outlier_thresh: float, optional
+            Suppress samples where the norm of the second-order differences of
+            the rotation is above `outlier_thresh` and interpolate the missing
+            values.
+
+        cutoff: float, optional
+            Frequency of a low-pass filter applied to linear and angular
+            velocity after the estimation as a fraction of the Nyquist
+            frequency.
+
+        mode: str, default "quaternion"
+            If "quaternion", compute the angular velocity from the quaternion
+            derivative. If "rotation_vector", compute the angular velocity from
+            the gradient of the axis-angle representation of the rotations.
+
+        allow_static: bool, default False
+            If True, return a zero velocity vector and None for timestamps if
+            the transform between this frame and the reference frame is static.
+            Otherwise, a `ValueError` will be raised.
+
+        return_timestamps: bool, default False
+            If True, also return the timestamps of the lookup.
+
+        Returns
+        -------
+        angular: numpy.ndarray, shape (N, 3)
+            Angular velocity of moving frame wrt reference frame, represented
+            in representation frame.
+
+        timestamps: each numpy.ndarray
+            Timestamps of the angular velocity.
+        """
+        try:
+            reference = _resolve_rf(reference or self.parent)
+            represent_in = _resolve_rf(represent_in or self.parent)
+        except TypeError:
+            raise ValueError(f"Frame {self.name} has no parent frame")
+
+        _, rotation, timestamps = self.lookup_transform(reference)
+
+        if timestamps is None:
+            if allow_static:
+                return np.zeros(3), None
+            else:
+                raise ValueError(
+                    "Velocity cannot be estimated for static transforms"
+                )
+
+        angular = _estimate_angular_velocity(
+            rotation,
+            timestamps,
+            cutoff=cutoff,
+            mode=mode,
+            outlier_thresh=outlier_thresh,
+        )
+
+        # angular velocity is represented in moving frame after estimation
+        angular, angular_ts = self.transform_vectors(
+            angular,
+            represent_in,
+            timestamps=timestamps,
+            return_timestamps=True,
+        )
+
+        if return_timestamps:
+            return angular, angular_ts
+        else:
+            return angular
 
     def register(self, update=False):
         """ Register this frame in the registry.
